

/*
**	DBLLIST, 1999-2001. Jinhong Kim, cheriz@empal.com
**
**  THIS IS COMPLETELY FREE!  ANYONE WILL BE PERMITTED TO MODIFY THE CODE FOR HIS/HER OWN PURPOSE.
**  (THE PURPOSE INCLUDES THE COMMERCIAL)
**  BUT, YOU SHOULD KNOW THAT I'M FREE FROM ANY-RESPONSIBILITIES FOR- 
**  ANY SYSTEM RELATED ERRORS MAY OCCUR WHILE USING THE CODE.
**  FREE YOUR MIND.
*/


// * doubly linked list template
// * it's STL-like. simple and convenient
// * support stack and queue usage
// * support arbitrary data type
// * support random access
// * support reverse
// * will support if I need: qsort, merge



#ifndef _LINKEDLIST_H
#define _LINKEDLIST_H

#define BUILD "1"

#pragma message (__FILE__ " [dbllist] cheriz@empal.com, build "BUILD)

#undef BUILD


#include <windows.h>


#define DECL_TEMPLATE	template <class ValueType>
#define DECL_CLASS		LinkedList<ValueType>
#define DECL_PROTOC		DECL_CLASS DECL_CLASS	// returnable class
#define DECL_PROTO		DECL_CLASS



DECL_TEMPLATE
class LinkedList
{
public:
	class Node
	{
		friend LinkedList;

	private:
		ValueType	v;
		Node*		ptr;
		Node*		prev;
		Node*		next;
		int			value_idx;

	public:
		Node()
		{
			ptr	= NULL;
			prev = NULL;
			next = NULL;
			value_idx = 0;
		}
		Node(Node& n)
		{
			Node();
			v = n.v;
			ptr = n.ptr;
			prev = n.prev;
			next = n.next;
		}
		Node(Node* n)
		{
			Node();
			v = n->v;
			ptr = n->ptr;
			prev = n->prev;
			next = n->next;
		}
		~Node() {}


		Node* Prev() { return prev; }
		Node* Next() { return next; }

		Node& operator[](int x)
		{
			if (value_idx == x)
				return (*ptr);
			
			int diff = x - value_idx;
			static int i;
			
			if (diff > 0)
			{
				if (1 == diff) return ++(*this);
				if (2 == diff) return ++(++(*this));
				if (3 == diff) return ++(++(++(*this)));
				for (i=0; i<diff; i++) (*this)++;
			}
			else
			{
				if (-1 == diff) return --(*this);
				if (-2 == diff) return --(--(*this));
				if (-3 == diff) return --(--(--(*this)));
				for (i=0; i<-diff; i++) (*this)--;
			}
					
			return (*this);
		}

		// returns the node's real address; to use in list operation
		Node* operator&() { return ptr;	}
		// copies the right node. and this is extendable.  i.e.) a = b = c;
		Node& operator=(Node& n)
		{
			ptr = n.ptr;
			v = n.v;
			next = n.next;
			prev = n.prev;
			return (*this);
		}
		Node& operator=(Node* n)
		{
			ptr = n;
			v = n->v;
			next = n->next;
			prev = n->prev;
			return (*this);
		}
		BOOL operator!=(Node* n) { return !((*this) == n); }
		BOOL operator==(Node* n) { return ( ptr == n->ptr ); }
		BOOL operator!=(ValueType value) { return !( (*this) == value ); }
		BOOL operator==(ValueType value) { return ( (ptr->v == value) ); }
		Node& operator++() { (*this)++; return (*this); }
		Node& operator++(int)
		{
			if (NULL != next)
			{
				value_idx++;
				ptr = next->ptr;
				v = next->v;
				prev = next->prev;
				next = next->next;
			}
			return (*this);
		}
		Node& operator--() { (*this)--; return (*this); }
		Node& operator--(int)
		{
			if (NULL != prev)
			{
				value_idx--;
				ptr = prev->ptr;
				v = prev->v;
				next = prev->next;
				prev = prev->prev;
			}
			return (*this);
		}
		ValueType operator*()	{ return ptr->v; }
		// let the real item's value of the l-valued node
		ValueType operator=(ValueType value) { ptr->v = value; return (*this); }
	};

	LinkedList();
	~LinkedList();

	// PushFront
	Node* PushFront(ValueType value);

	// insert value after the last
	Node* PushBack(ValueType value);

	void PushBefore(Node* n, ValueType value);
	void PushAfter(Node* n, ValueType value);

	// stack style pop
	ValueType Pop();

	// queue style pop
	ValueType QPop();

	Node* Front();
	Node* Back();

	ValueType GetValue(Node* n);
	void SetValue(ValueType value);

	BOOL IsEmpty();

	BOOL Remove(Node* n);
	BOOL RemoveFront();
	BOOL RemoveBack();
	void RemoveAll();

	int GetTotalNum();
	void Reverse();

private:
	Node*	pFirst;
	Node*	pLast;
	int		value_counter;

private:
	Node* _makeitem(ValueType value);
	void _chkvalidate();
};


DECL_TEMPLATE
DECL_PROTO::LinkedList()
{
	pFirst = NULL;
	pLast = NULL;
	value_counter = 0;
}


DECL_TEMPLATE
DECL_PROTO::~LinkedList()
{
	RemoveAll();
}


DECL_TEMPLATE
DECL_PROTO::Node* DECL_CLASS::_makeitem(ValueType value)
{
	Node* item;

	item = new Node;

	item->ptr = item;
	item->v = value;
	item->prev = NULL;
	item->next = NULL;

	return item;
}


DECL_TEMPLATE
DECL_PROTO::Node* DECL_CLASS::PushFront(ValueType value)
{
	Node* item = _makeitem(value);

	if (NULL == pFirst)
	{
		pFirst = item;
		pLast = item;
	}
	else
	{
		item->next = pFirst;
		pFirst->prev = item;
		pFirst = item;
	}

	value_counter++;

	return item;
}


DECL_TEMPLATE
DECL_PROTO::Node* DECL_CLASS::PushBack(ValueType value)
{
	Node* item = _makeitem(value);

	if (NULL == pFirst)
	{
		pFirst = item;
		pLast = item;
	}
	else
	{
		pLast->next = item;
		item->prev = pLast;
		pLast = item;
	}

	value_counter++;
	return item;
}


DECL_TEMPLATE
void DECL_CLASS::PushBefore(Node* n, ValueType value)
{
	Node* item = _makeitem(value);

	if (n == pFirst || NULL == pFirst)
	{
		PushFront(value);
		return;
	}

	item->next = n;
	item->prev = n->prev;
	if (NULL != n->prev)
		n->prev->next = item;
	n->prev = item;
}

DECL_TEMPLATE
void DECL_CLASS::PushAfter(Node* n, ValueType value)
{
	Node* item = _makeitem(value);

	if (n == pLast || NULL == pFirst)
	{
		PushBack(value);
		return;
	}

	item->prev = n;
	item->next = n->next;
	n->next = item;
	if (NULL != n->next)
		n->next->prev = item;
}


DECL_TEMPLATE
void DECL_CLASS::_chkvalidate()
{
	if (0 == value_counter)
	{
		pFirst = NULL;
		pLast = NULL;
	}
}


DECL_TEMPLATE
BOOL DECL_CLASS::RemoveFront()
{
	Node* tmp;

	if (NULL == pFirst)
		return FALSE;

	if (pFirst != pLast)
		pFirst->next->prev = NULL;

	tmp = pFirst->next;

	delete pFirst;
	pFirst = tmp;
	value_counter--;

	_chkvalidate();
	return TRUE;
}


DECL_TEMPLATE
BOOL DECL_CLASS::RemoveBack()
{
	Node* tmp;

	if (NULL == pFirst)
		return FALSE;

	if (pFirst != pLast)
		pLast->prev->next = NULL;

	tmp = pLast->prev;
	delete pLast;
	pLast = tmp;
	value_counter--;

	_chkvalidate();
	return TRUE;
}



DECL_TEMPLATE
BOOL DECL_CLASS::Remove(Node* n)
{
	if (NULL == pFirst)
		return FALSE;

	if (NULL == n)
		return FALSE;

	if (n == pFirst)
		return RemoveFront();

	if (n == pLast)
		return RemoveBack();

	if (NULL != n->prev)
		n->prev->next = n->next;

	if (NULL != n->next)
		n->next->prev = n->prev;

	delete n;
	value_counter--;

	_chkvalidate();
	return TRUE;
}

DECL_TEMPLATE
void DECL_CLASS::RemoveAll()
{
	Node* tmp;

	if (NULL == pFirst)
		return;

	while (1)
	{
		tmp = pFirst;
		pFirst = pFirst->next;

		if (NULL == pFirst)
		{
			delete tmp;
			break;
		}

		pFirst->prev = NULL;
		delete tmp;
	}

	pFirst = NULL;
	pLast = NULL;
	value_counter = 0;
}


DECL_TEMPLATE
DECL_PROTO::Node* DECL_CLASS::Front()
{
	return pFirst;
}

DECL_TEMPLATE
DECL_PROTO::Node* DECL_CLASS::Back()
{
	return pLast;
}

DECL_TEMPLATE
ValueType DECL_CLASS::GetValue(Node* n)
{
	return n->v;
}

DECL_TEMPLATE
void DECL_CLASS::SetValue(ValueType value)
{
	n->v = value;
}

DECL_TEMPLATE
BOOL DECL_CLASS::IsEmpty()
{
	return NULL == pFirst;
}


DECL_TEMPLATE
ValueType DECL_CLASS::QPop()
{
	ValueType v;

	if (NULL != pFirst)
	{
		v = GetValue(pFirst);
		RemoveFront();
	}

	return v;
}


DECL_TEMPLATE
ValueType DECL_CLASS::Pop()
{
	ValueType v;

	if (NULL != pFirst)
	{
		v = GetValue(pLast);
		RemoveBack();
	}

	return v;
}


DECL_TEMPLATE
int DECL_CLASS::GetTotalNum()
{
	return value_counter;
}


DECL_TEMPLATE
void DECL_CLASS::Reverse()
{
	Node*	n = Front();
	Node*	tmp;

	if (NULL == pFirst)
		return;

	while(1)
	{
		tmp = n->prev;
		n->prev = n->next;
		n->next = tmp;
		if (n == pLast) break;
		n = n->prev;	// prev has 'next' ptr
	}

	tmp = pFirst;
	pFirst = pLast;
	pLast = tmp;
}


#endif

