

/*
**	W3D basic.  Jinhong Kim, cheriz@empal.com
**
**  THIS IS COMPLETELY FREE!  ANYONE WILL BE PERMITTED TO MODIFY THE CODE FOR HIS/HER OWN PURPOSE.
**  (THE PURPOSE INCLUDES THE COMMERCIAL)
**  BUT, YOU SHOULD KNOW THAT I'M FREE FROM ANY-RESPONSIBILITIES FOR- 
**  ANY SYSTEM RELATED ERRORS MAY OCCUR WHILE USING THE CODE.
**  FREE YOUR MIND.
*/



#include <windows.h>
#include <gl/gl.h>
#include <gl/glu.h>
#include <gl/glaux.h>
#include <math.h>
#include <time.h>
#include <stdio.h>

#include "dbllist.h"
#include "w3Dcore.h"
#include "w3dmath.h"


#pragma warning (disable : 4710)	// not inlined warning; This will keep on.


//////////////////////////////////////////////////////////////////////////////
// INTERNAL ONLY FUNCTION PROTOTYPE FORWARD DECLARATIONS

static BOOL classify_attrib(w3DObject *pObj);


//////////////////////////////////////////////////////////////////////////////
// IMPLEMENTATION: w3DObject

w3DObject::w3DObject()
{
	ZeroMemory((void*)_name, NAMESIZE);
}

w3DObject::~w3DObject()
{
}

char* w3DObject::GetName()
{
	return _name;
}

void w3DObject::SetName(char *Name)
{
	CopyMemory((void*)_name, (void*)Name, NAMESIZE);
}



//////////////////////////////////////////////////////////////////////////////
// IMPLEMENTATION: w3DMalloc

w3DMalloc::w3DMalloc() 
{ 
	pMalloc = NULL; 
	pHeadDW = NULL; 
	pTailDW = NULL; 
	Size = 0;
	ActualSize = 0;
	Used = FALSE;
}

w3DMalloc::~w3DMalloc() 
{
	if (Used)
		Free();
}

void w3DMalloc::Free()
{
	if (pHeadDW == NULL)
	{
#ifdef	W3DDEBUG
		w3DDebugMsg("Attempt to deallocate non-allocated memory!\n");
#endif
		return;
	}
	
	if (!Used)
	{
#ifdef	W3DDEBUG
		w3DDebugMsg("Already deallocated block!\n");
#endif
		return;
	}
	
	if (*pTailDW != 0xDCBA4321)
	{
#ifdef	W3DDEBUG
		w3DDebugMsg("This memory block was currupted! Cannot be freed.\n");
#endif
		return;
	}
	
	Used = FALSE;
	
	VirtualUnlock(pMalloc, ActualSize);
	VirtualFree(pMalloc, ActualSize, MEM_DECOMMIT);
	
	delete this;
}

void* w3DMalloc::Malloc(DWORD len) 
{ 
	char *guide1; // for easy pointer calculation
	void *pUserMemory;
	
	if (len == 0)
	{
#ifdef	W3DDEBUG
		w3DDebugMsg("ignored 0 byte allocation!\n");
#endif
		return NULL;
	}
	
#define W3DMEMHEADSIZE	sizeof(DWORD) * 2
#define W3DMEMTAILSIZE	sizeof(DWORD)
	
	Size = len;
	ActualSize = Size + W3DMEMHEADSIZE + W3DMEMTAILSIZE;
	
	pMalloc = VirtualAlloc(NULL, ActualSize,
		MEM_COMMIT, PAGE_READWRITE);
	VirtualLock(pMalloc, ActualSize);
	
	guide1 = (char*)pMalloc;
	pHeadDW = (DWORD*)guide1;
	pUserMemory = (void*)&guide1[W3DMEMHEADSIZE];
	
	guide1 = (char*)pUserMemory;
	pTailDW = (DWORD*)&guide1[Size];
	
	// write head signature
	pHeadDW[0] = 0x1234ABCD;
	
	// write 'this' pointer
	pHeadDW[1] = (DWORD)this;
	
	pTailDW[0] = 0xDCBA4321;
	
	Used = TRUE;
	return pUserMemory; 
}

void w3DMalloc::Clean()
{
	char *guide1; // for easy pointer calculation
	void *pUserMemory;

	if (!Used)
	{
		w3DDebugMsg("Attempt to clean not yet allocated memory block!\n");
		return;
	}

	guide1 = (char*)pMalloc;
	pHeadDW = (DWORD*)guide1;
	pUserMemory = (void*)&guide1[W3DMEMHEADSIZE];

	ZeroMemory(pUserMemory, Size);
}

#undef W3DMEMHEADSIZE
#undef W3DMEMTAILSIZE



w3DMalloc* w3DMalloc::Bind(void* ptr)
{
	DWORD *pH;
	
	pH = (DWORD*)ptr;
	pH--; pH--;
	
	if (*pH != 0x1234ABCD)
	{
#ifdef	W3DDEBUG
		w3DDebugMsg("This memory was not allocated by w3DMalloc!\n");
#endif
		return NULL;
	}
	
	pH++;
	return (w3DMalloc*)*pH;
}

void w3DMalloc::Free(void* ptr)
{
	DWORD *pH;
	
	pH = (DWORD*)ptr;
	pH--; pH--;
	
	if (*pH != 0x1234ABCD)
	{
#ifdef	W3DDEBUG
		w3DDebugMsg("This memory was not allocated by w3DMalloc!\n");
#endif
		return;
	}
	
	pH++;
	((w3DMalloc*)*pH)->Free();
}




//////////////////////////////////////////////////////////////////////////////
// IMPLEMENTATION: w3DGroup

w3DGroup::w3DGroup() 
{
}

w3DGroup::~w3DGroup() 
{
	Empty();
}

W3DGROUPNODETYPE* w3DGroup::Add(w3DObject* item)
{
	return list.PushBack(item);
}

void w3DGroup::FindNext()
{
	node++;
}

void w3DGroup::FindFirst()
{
	node = list.Front();
}

w3DObject* w3DGroup::GetData()
{
	return *node;
}

W3DGROUPNODETYPE& w3DGroup::GetCurNode()
{
	return node;
}

void w3DGroup::SetCurNode(W3DGROUPNODETYPE* pnode)
{
	node = pnode;
}

BOOL w3DGroup::DeleteFirst()
{
	BOOL b;
	b = list.RemoveFront();
	FindFirst();
	return b;
}

void w3DGroup::Empty()
{
	list.RemoveAll();
}

void w3DGroup::SetParent(w3DGroup* pParent_)
{
	pParent = pParent_;
}

w3DGroup* w3DGroup::GetParent()
{
	return pParent;
}

int w3DGroup::GetTotalNum()
{
	return list.GetTotalNum();
}


//////////////////////////////////////////////////////////////////////////////
// IMPLEMENTATION: w3DModel

w3DModelBase::w3DModelBase()
{
	attr = NULL;
}

w3DModelBase::~w3DModelBase() 
{
}

void w3DModelBase::Render() 
{
}

w3DGroup* w3DModelBase::GetAttrib()
{
	return attr;
}

void w3DModelBase::SetAttrib(w3DGroup* attr)
{
	this->attr = attr;
}



//////////////////////////////////////////////////////////////////////////////
// IMPLEMENTATION: w3DSystem

const GLfloat w3DSystem::PI				= 3.14159265359f;
const GLfloat w3DSystem::TORAD			= 0.017453292519944f;
const GLfloat w3DSystem::TODEG			= 57.295780490443f;
const char* const w3DSystem::Version	= "0.0.1";
HWND	w3DSystem::hWnd					= NULL;
HDC		w3DSystem::hDC					= NULL;
HGLRC	w3DSystem::hRC					= NULL;
BOOL	w3DSystem::bInitialized			= FALSE;
w3DCamera* w3DSystem::ActiveCamera		= NULL;
float	w3DSystem::Aspect				= 0.0f;
GLfloat w3DSystem::bkColor[4]			= {0.0f, 0.0f, 0.0f, 1.0f};
BOOL	w3DSystem::cull_face			= TRUE;
BOOL	w3DSystem::PolygonMode			= 0;

GLfloat w3DSystem::ambientLight[4] = {0.15f, 0.15f, 0.15f, 1.0f};
GLfloat w3DSystem::diffuseLight[4] = {1.0f, 1.0f, 1.0f, 1.0f};
GLfloat w3DSystem::specularLight[4] = {1.0f, 1.0f, 1.0f, 1.0f};
GLfloat w3DSystem::lightPos[4] = {-1000.0f, 1000.0f, 1000.0f, 1.0f};

BYTE w3DSystem::DepthBufferSize = 16;
BYTE w3DSystem::ColorDepth = 32;


w3DSystem::w3DSystem()
{
	SetBackGroundColor(0.0f, 0.0f, 0.0f, 1.0f);
	SetAmbientLight(0.15f, 0.15f, 0.15f, 1.0f);
	SetDiffuseLight(1.0f, 1.0f, 1.0f, 1.0f);
	SetSpecularLight(1.0f, 1.0f, 1.0f, 1.0f);
	SetLightPos(-1000.0f, 1000.0f, 1000.0f, 1.0f);
	SetColorDepth(32);
	SetDepthBufferSize(16);

	pModeler = NULL;
	apply_attr = NULL;

	stop = FALSE;
}

w3DSystem::~w3DSystem()
{
	wglDeleteContext(w3DSystem::hRC);
	::ReleaseDC(w3DSystem::hWnd, w3DSystem::hDC);
}

// THIS IS A STATIC MEMBER FUNCTION
void w3DSystem::SetAmbientLight(GLfloat r, GLfloat g, GLfloat b, GLfloat a)
{
	ambientLight[0] = r;
	ambientLight[1] = g;
	ambientLight[2] = b;
	ambientLight[3] = a;
}

// THIS IS A STATIC MEMBER FUNCTION
void w3DSystem::SetDiffuseLight(GLfloat r, GLfloat g, GLfloat b, GLfloat a)
{
	diffuseLight[0] = r;
	diffuseLight[1] = g;
	diffuseLight[2] = b;
	diffuseLight[3] = a;
}

// THIS IS A STATIC MEMBER FUNCTION
void w3DSystem::SetSpecularLight(GLfloat r, GLfloat g, GLfloat b, GLfloat a)
{
	specularLight[0] = r;
	specularLight[1] = g;
	specularLight[2] = b;
	specularLight[3] = a;
}

// THIS IS A STATIC MEMBER FUNCTION
void w3DSystem::SetLightPos(GLfloat x, GLfloat y, GLfloat z, GLfloat a)
{
	lightPos[0] = x;
	lightPos[1] = y;
	lightPos[2] = z;
	lightPos[3] = a;
}

// THIS IS A STATIC MEMBER FUNCTION
void w3DSystem::SetBackGroundColor(GLfloat r, GLfloat g, GLfloat b, GLfloat a)
{
	bkColor[0] = r;
	bkColor[1] = g;
	bkColor[2] = b;
	bkColor[3] = a;
}

// THIS IS A STATIC MEMBER FUNCTION
void w3DSystem::SetCullFace(BOOL bi)
{
	cull_face = bi;
}

// A STATIC MEMBER FUNCTION
void w3DSystem::SetPolygonMode(BOOL mode)
{
	PolygonMode = mode;
}

void w3DSystem::AdjustCSInfo(CREATESTRUCT &cs)
{
	cs.style |= (WS_CLIPCHILDREN | WS_CLIPSIBLINGS | CS_OWNDC);
}

void w3DSystem::ApplyAttrib(w3DGroup *attr)
{
	apply_attr = attr;
}

void w3DSystem::SetupRC()
{
	if (cull_face)
	{
		glEnable(GL_CULL_FACE);
		glCullFace(GL_BACK);
	}
	else
		glDisable(GL_CULL_FACE);

	glEnable(GL_DEPTH_TEST);
	glFrontFace(GL_CCW);

	if (PolygonMode == 0)
		glPolygonMode(GL_FRONT, GL_FILL);
	else
		glPolygonMode(GL_FRONT, GL_LINE);

	glShadeModel(GL_SMOOTH);
	glEdgeFlag(TRUE);


//	glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
//	glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
//	glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);
//	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

//	glEnable(GL_LINE_SMOOTH);
//	glEnable(GL_POINT_SMOOTH);
//	glEnable(GL_POLYGON_SMOOTH);


	// light setting
	glEnable(GL_LIGHTING);

	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambientLight);
//	GLfloat lv = 0.0f;
//	glLightModelfv(GL_LIGHT_MODEL_LOCAL_VIEWER, &lv);
//	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
	glLightfv(GL_LIGHT0, GL_AMBIENT, ambientLight);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuseLight);
	glLightfv(GL_LIGHT0, GL_SPECULAR, specularLight);
	glEnable(GL_LIGHT0);
	
	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);

	GLfloat specref[] = {1.0f, 1.0f, 1.0f, 1.0f};
	glMaterialfv(GL_FRONT, GL_SPECULAR, specref);
	glMateriali(GL_FRONT, GL_SHININESS, 128);

	glClearColor(bkColor[0], bkColor[1], bkColor[2], 1.0f);
}

// STATIC FUNCTION
void w3DSystem::SetColorDepth(BYTE cd)
{
	ColorDepth = cd;
}

// STATIC FUNCTION
void w3DSystem::SetDepthBufferSize(BYTE db)
{
	DepthBufferSize = db;
}

void w3DSystem::Init(HWND hWnd)
{
	LONG nPixelFormat;

	w3DSystem::hWnd = hWnd;
	w3DSystem::hDC = ::GetDC(hWnd);

	if (NULL == w3DSystem::hWnd || NULL == w3DSystem::hDC)
		return;


	static PIXELFORMATDESCRIPTOR pfd = {
		sizeof(PIXELFORMATDESCRIPTOR),	// Size of this structure
		1,								// Version of this structure	
		PFD_DRAW_TO_WINDOW |			// Draw to Window (not to bitmap)
		PFD_SUPPORT_OPENGL |			// Support OpenGL calls in window
		PFD_DOUBLEBUFFER,				// Double buffered mode
		PFD_TYPE_RGBA,					// RGBA Color mode
		ColorDepth,						// Color depth
		0,0,0,0,0,0,					// Not used to select mode
		0,0,							// Not used to select mode
		0,								// Accumulation Buffer
		0,
		0,
		0,
		0,
		DepthBufferSize,				// Size of depth buffer
		0,								// Stencil
		0,								// Not used to select mode
		PFD_MAIN_PLANE,					// Draw in main plane
		0,								// Not used to select mode
		0,0,0 };						// Not used to select mode

	// Choose a pixel format that best matches that described in pfd
	nPixelFormat = ChoosePixelFormat(w3DSystem::hDC, &pfd);

	// Set the pixel format for the device context
	if (!SetPixelFormat(hDC, nPixelFormat, &pfd))
	{
		w3DMsgBox("OpenGL PixelFormat could not be set!\n");
		return;
	}

	// Create the rendering context
	w3DSystem::hRC = wglCreateContext(w3DSystem::hDC);
	w3DSystem::bInitialized = TRUE;

	w3DContext(TRUE);

	// setup rendering context
	SetupRC();

	// display GL informations
#ifdef W3DDEBUG
	w3DErrorTest("After SetupRC()");

	w3DDebugMsg("--- W3D Informations ---\n");
	w3DDebugMsg("System Version: %s\n", Version);
	w3DDebugMsg("--- OpenGL Informations ---\n");
	w3DDebugMsg("Vendor: %s\n", glGetString(GL_VENDOR));
	w3DDebugMsg("Renderer: %s\n", glGetString(GL_RENDERER));
	w3DDebugMsg("Version: %s\n", glGetString(GL_VERSION));

	LONG i, slen;
	const unsigned char *sptr = glGetString(GL_EXTENSIONS);
	char buffer[2500];

	slen = strlen((const char*)sptr);
	CopyMemory((void*)buffer, (void*)sptr, slen);

	for (i=0; i<slen; i++)
		if (buffer[i] == ' ')
			buffer[i] = '\n';

	w3DDebugMsg("Extensions:\n");
	w3DDebugMsg("%s\n\n", buffer);
	w3DDebugMsg("\n");
#endif

	w3DContext(FALSE);

	// do Resize first
	RECT r;
	::GetClientRect(hWnd, &r);
	Resize(r.right, r.bottom);
}


void w3DSystem::Resize(GLsizei w, GLsizei h)
{
	if (!w3DSystem::bInitialized)
		return;

	if(0 == h)
		h = 1;

	w3DContext(TRUE);

	// Set Viewport to window dimensions
    glViewport(0, 0, w, h);

	// Reset coordinate system
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	w3DSystem::Aspect = (GLfloat)w/(GLfloat)h;
	gluPerspective(pCamera->GetFOV(), Aspect, pCamera->GetNearDist(), pCamera->GetFarDist());

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	w3DContext(FALSE);
}

w3DCamera* w3DSystem::GetActiveCamera()
{
	return ActiveCamera;
}

void w3DSystem::PrepareRendering()
{
	// prepare rendering
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT /* | GL_ACCUM_BUFFER_BIT */);

	// force to apply the attributes now!
	if (apply_attr)
	{
		w3DApplyAttrib(apply_attr);
		apply_attr = NULL;
	}

	// positioning the light
	glLightfv(GL_LIGHT0, GL_POSITION, lightPos);
}

void w3DSystem::RenderScene()
{
	PrepareRendering();

	// apply camera transform
	pCamera->Apply();

	// render the scene
	pModeler->Process();
}

void w3DSystem::Process()
{
	if (!w3DSystem::bInitialized)
	{
#ifdef	W3DDEBUG
		w3DDebugMsg(W3D_ERR_NOTINITIALIZED);
#endif
		return;
	}

	if (!stop)
	{
		w3DContext(TRUE);
		RenderScene();
		w3DContext(FALSE);
	}
}

void w3DSystem::Swap()
{
	w3DFlushAndSwap();
}

void w3DSystem::SetModeler(w3DModeler *pMod) 
{ 
	pModeler = pMod; 
}

void w3DSystem::SetActiveCamera(w3DCamera *pCam) 
{ 
	pCamera = pCam; 
	ActiveCamera = pCamera;
}

void w3DSystem::Stop() 
{ 
	stop = TRUE; 
}

void w3DSystem::Resume() 
{ 
	stop = FALSE; 
}



//////////////////////////////////////////////////////////////////////////////
// IMPLEMENTATION: w3DIntegration

w3DIntegration::w3DIntegration()
{
}

w3DIntegration::~w3DIntegration() 
{
}

void w3DIntegration::Init(HWND hwnd)
{
	sys.SetActiveCamera(&cam);
	sys.Init(hwnd);
	sys.SetModeler(&mod);
}



//////////////////////////////////////////////////////////////////////////////
// IMPLEMENTATION: w3DModeler

w3DModeler::w3DModeler()
{
	model = NULL;
}

w3DModeler::~w3DModeler() 
{
}

void w3DModeler::Process()
{
	w3DGroup* pScan;
	w3DModelBase* pModCur;
	DWORD mt;

	// returns, if there are no primitives registered.
	if (model == NULL)
		return;

	pScan = model;
	pScan->FindFirst();


	for (int i = 0; i < pScan->GetTotalNum(); i++)
	{
		pModCur = (w3DModelBase*)pScan->GetData();

		mt = pModCur->GetType();

		if (mt == W3D_CLS_MODEL || mt == W3D_CLS_IMODEL)
		{
			w3DApplyAttrib(pModCur->GetAttrib());
			pModCur->Render();
		}

		pScan->FindNext();
	}
}

void w3DModeler::SetModel(w3DGroup* mod)
{
	this->model = mod;
}




//////////////////////////////////////////////////////////////////////////////
// IMPLEMENTATION: w3DObjSphere

w3DObjSphere::w3DObjSphere(GLdouble r)
{
	radius = r;
}

w3DObjSphere::~w3DObjSphere()
{
}

void w3DObjSphere::Render()
{
	auxSolidSphere(radius);
}



//////////////////////////////////////////////////////////////////////////////
// IMPLEMENTATION: w3DObjCylinder

w3DObjCylinder::w3DObjCylinder(GLdouble r, GLdouble h)
{
	radius = r;
	height = h;
}

w3DObjCylinder::~w3DObjCylinder()
{
}

void w3DObjCylinder::Render()
{
	auxSolidCylinder(radius, height);
}


//////////////////////////////////////////////////////////////////////////////
// IMPLEMENTATION: w3DCamera

w3DCamera::w3DCamera(float x, float y, float z, float _pitch, float _yaw, float _roll) 
{
	Set(w3DPoint3(x, y, z), w3DPoint3(x, y, z - 1.0f), w3DVector3(0, 1, 0));
	Pitch(_pitch);
	Yaw(_yaw);
	Roll(_roll);
	
	ln = n;
	lu = u;
	lv = v;
	le = eye;
	
	fieldofview = 45.0f;
	nearDist = 15.0f;
	farDist = 5000.0f;
}

w3DCamera::w3DCamera()
{
	Set(w3DPoint3(), w3DPoint3(0, 0, -1), w3DVector3(0, 1, 0));
	
	ln = n;
	lu = u;
	lv = v;
	le = eye;

	fieldofview = 45.0f;
	nearDist = 15.0f;
	farDist = 5000.0f;
}

w3DCamera::~w3DCamera()
{}


void w3DCamera::Apply()
{
	static float m[16];

	w3DVector3 eVec(eye.x, eye.y, eye.z);
	m[0] = u.x;  m[4] = u.y;  m[8] = u.z;  m[12] = -eVec.Dot(u);
	m[1] = v.x;  m[5] = v.y;  m[9] = v.z;  m[13] = -eVec.Dot(v);
	m[2] = n.x;  m[6] = n.y;  m[10] = n.z; m[14] = -eVec.Dot(n);
	m[3] = 0;    m[7] = 0;    m[11] = 0;   m[15] = 1.0;

	glMatrixMode(GL_MODELVIEW);
	glLoadMatrixf(m);
}

void w3DCamera::Set(w3DPoint3 e, w3DPoint3 l, w3DVector3 up)
{
	eye.Set(e);
	n.Set(eye.x - l.x, eye.y - l.y, eye.z - l.z);
	u.Set(up.Cross(n));
	n.Norm();
	u.Norm();
	v.Set(n.Cross(u));
}


void w3DCamera::Set(float x, float y, float z, float _pitch, float _yaw, float _roll) 
{
	Set(w3DPoint3(x, y, z), w3DPoint3(x, y, z - 1.0f), w3DVector3(0, 1, 0));
	Pitch(_pitch);
	Yaw(_yaw);
	Roll(_roll);
	
	ln = n;
	lu = u;
	lv = v;
	le = eye;
}

void w3DCamera::ResetAngle()
{
	n = ln;
	u = lu;
	v = lv;
}

void w3DCamera::Roll(float angle)
{
	float cs = (float)cos(-angle * w3DSystem::TORAD);
	float sn = (float)sin(-angle * w3DSystem::TORAD);

	w3DVector3 t = u;

	u.Set(cs*t.x - sn*v.x, cs*t.y - sn*v.y, cs*t.z - sn*v.z);
	v.Set(sn*t.x + cs*v.x, sn*t.y + cs*v.y, sn*t.z + cs*v.z);
}

void w3DCamera::Pitch(float angle)
{
	float cs = (float)cos(angle * w3DSystem::TORAD);
	float sn = (float)sin(angle * w3DSystem::TORAD);

	w3DVector3 t = n;

	n.Set(cs*t.x - sn*v.x, cs*t.y - sn*v.y, cs*t.z - sn*v.z);
	v.Set(sn*t.x + cs*v.x, sn*t.y + cs*v.y, sn*t.z + cs*v.z);
}

void w3DCamera::Yaw(float angle)
{
	float cs = (float)cos(angle * w3DSystem::TORAD);
	float sn = (float)sin(angle * w3DSystem::TORAD);

	w3DVector3 t = n;

	n.Set(cs*t.x - sn*u.x, cs*t.y - sn*u.y, cs*t.z - sn*u.z);
	u.Set(sn*t.x + cs*u.x, sn*t.y + cs*u.y, sn*t.z + cs*u.z);
}

void w3DCamera::Slide(float dU, float dV, float dN)
{
	eye.x += dU*u.x + dV*v.x + dN*n.x;
	eye.y += dU*u.y + dV*v.y + dN*n.y;
	eye.z += dU*u.z + dV*v.z + dN*n.z;
}

w3DPoint3 w3DCamera::GetPos()
{
	return eye;
}

void w3DCamera::SetParam(float vAng, float nearD, float farD)
{
	fieldofview = vAng;
	nearDist = nearD;
	farDist = farD;
}

float w3DCamera::GetFOV() 
{ 
	return fieldofview; 
}

float w3DCamera::GetNearDist() 
{ 
	return nearDist; 
}

float w3DCamera::GetFarDist() 
{ 
	return farDist; 
}



//////////////////////////////////////////////////////////////////////////////
// MISCELLANEOUS IMPLEMENTATIONS

void w3DContext(BOOL _ctx)
{
	if (!w3DSystem::bInitialized)
	{
#ifdef	W3DDEBUG
		w3DDebugMsg(W3D_ERR_NOTINITIALIZED);
#endif
		return;
	}

	if (_ctx)
		wglMakeCurrent(w3DSystem::hDC, w3DSystem::hRC);
	else
		wglMakeCurrent(NULL, NULL);
}

void w3DFlushAndSwap()
{
	w3DContext(TRUE);
	glFlush();
	SwapBuffers(w3DSystem::hDC);
	w3DContext(FALSE);
}

void w3DMsgBox(const char *fmt, ...)
{
	char		text[256];									// Holds Our String
	va_list		ap;											// Pointer To List Of Arguments

	if (fmt == NULL)										// If There's No Text
		return;												// Do Nothing

	va_start(ap, fmt);										// Parses The String For Variables
	    vsprintf(text, fmt, ap);							// And Converts Symbols To Actual Numbers
	va_end(ap);												// Results Are Stored In Text
	
	::MessageBox(NULL, text, "W3D System", MB_OK | MB_ICONERROR);
}

#pragma warning (disable : 4100)
void w3DDebugMsg(const char *fmt, ...)
{
#ifdef W3DDEBUG
	char		text[2560];									// Holds Our String
	va_list		ap;											// Pointer To List Of Arguments

	if (fmt == NULL)										// If There's No Text
		return;												// Do Nothing

	va_start(ap, fmt);										// Parses The String For Variables
	    vsprintf(text, fmt, ap);							// And Converts Symbols To Actual Numbers
	va_end(ap);												// Results Are Stored In Text

	OutputDebugString(text);
#endif
}
#pragma warning (default : 4100)


void w3DErrorTest(char* who)
{
	LONG i, error;
	
	for (i=0; i<6; i++)
	{
		error = glGetError();

		w3DDebugMsg("%s: %s\n", who, gluErrorString(error));

		if (error == GL_NO_ERROR)
			break;
	}
}


void w3DApplyAttrib(w3DGroup* pAttrGrp)
{
	// if there are no attributes...
	if (pAttrGrp == NULL)
		return;

	pAttrGrp->FindFirst();
	
	for (int i = 0; i < pAttrGrp->GetTotalNum(); i++)
	{
		if (classify_attrib(pAttrGrp->GetData()) == FALSE)
			break;
		pAttrGrp->FindNext();
	}
}


//////////////////////////////////////////////////////////////////////////////
// INTERNAL ONLY FUNCTION IMPLEMENTATIONS

static BOOL classify_attrib(w3DObject *pObj)
{
	if (pObj == NULL)
		return FALSE;

	switch (pObj->GetType())
	{
		// will this be useful??? Hmm...
	case W3D_ATTRIB_LOADIDENTITY:
		glLoadIdentity();
		break;

	case W3D_ATTRIB_TRANSLATE:
		{
			static w3DTranslate *pTrans;
			pTrans = (w3DTranslate*)pObj;
			if (pTrans->apply)
				glTranslatef(pTrans->x, pTrans->y, pTrans->z);
		}
		break;

	case W3D_ATTRIB_ROTATE:
		{
			static w3DRotate *pRot;
			pRot = (w3DRotate*)pObj;
			if (pRot->apply)
				glRotatef(pRot->a, pRot->x, pRot->y, pRot->z);
		}
		break;

	case W3D_ATTRIB_SCALE:
		{
			static w3DScale *pScl;
			pScl = (w3DScale*)pObj;
			glScalef(pScl->x, pScl->y, pScl->z);
		}
		break;

	case W3D_ATTRIB_COLOR:
		{
			static w3DColor *pCol;
			pCol = (w3DColor*)pObj;
			glColor4fv(pCol->rgba);
		}
		break;
	}

	return TRUE;
}



#undef W3DDEBUG
